//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Emulation.cs.                                                    *
//* CLASS    ArtificialGenetics.cs. (Artificial Genetic emulator)             *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     11/17/2013                                                       *
//* REVISION X.0.0.8                                                          *
//* BYTES    25,929                                                           *
//* TIME     19:00                                                            *
//* PURPOSE  To do something.                                                 *
//* METHOD   By means of software simulation, model proto-humans.             *
//* DETAILS  Created in C# to model proto-humans.                             *
//* ENVIRON  Developed on a 64-bit Intel Core2Quad, 2.33GHz, with 4GB RAM.    *
//*          The Zotac m-ITX motherboard had board-based Nvidia 9300 graphics *
//*          acceleration. The Operating System was Microsoft (MS)            *
//*          Windows 8 Pro.  The MS Integrated Development Environment (IDE)  *
//*          MS Visual Studio 2010 Version 10.0.40219.1 SP1Rel and MS .NET    *
//*          Framework Version 4.0.30319 SP1Rel.                              *
//*                                                                           *
//*****************************************************************************
//

using System;

//
//*****************************************************************************
//

namespace Nursery
{

    //
    //*************************************************************************
    //

    public class ArtificialGenetics
    {

        //
        //*********************************************************************
        // Class constants, properties, and objects.

        #region Variables Dictionary

        // Constants.
        const int GENE_LENGTH = 8;                // count of alleles per gene.
        const int GENE_COUNT = 11;            // count of genes per chromosome.
        const int RANDOM_TYPE = 0;              // gene type is uniform random.
        const int GAUSSIAN_TYPE = 1;           // gene type is Gaussian normal.
        const double MALE_SKEW = 0.495F;      // likelihood of male chromosome.

        // Variables.
        private int [] gaussDistribution = null;              // Gaussian data.
        private static int rngSeed;             // this is the GLOBAL RNG seed.

        // Objects.

        private Random coin = new Random ( rngSeed );//random number generator.

        #endregion Variables Dictionary

        //
        //*********************************************************************
        //

        #region Constructor

        ///<summary>
        /// this is the class constructor.
        ///</summary>
        public ArtificialGenetics ( int newSeed )
        {
            rngSeed = newSeed;                     // copy the GLOBAL RNG seed.
            coin = new Random ( rngSeed );        // a random number generator.

            // the agent Gauss Distribution data set will be built.
            this.gaussDistribution = makeGaussianDistribution ();
        }

        #endregion Constructor

        //
        //*********************************************************************
        //

        #region Methods

        /// <summary>
        /// this method randomly assigns a special characteristic to a
        /// chromosome.  It defines the sex (M/F) of the bearer.
        /// </summary>
        public String assignPhenotype ()
        {
            String g = new String ( '0', GENE_LENGTH ); // gene composed of 0s.
            char [] gene = g.ToCharArray ();          // converted to an array.

            // a persons gender is set according to probability.
            char gender = '0';                      // assume this will be "F".
            if ( coin.NextDouble () < MALE_SKEW )   // set this person's gender.
                gender = '1';                      // select either 0=F or 1=M.
            else
                gender = '0';                      // select either 0=F or 1=M.


            // bits in gene are counted from right to left as n - 0.  Insert
            // the sex bit into the lsb position.
            gene [ GENE_LENGTH - 1 ] = gender;        // insert sex bit in lsb.

            return CharArrayToString ( gene );
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method randomly assigns a special characteristic to a
        /// chromosome.  It defines the sex (M/F) of the bearer.
        /// </summary>
        public String assignPhenotype ( String setSex )
        {
            String g = new String ( '0', GENE_LENGTH ); // gene composed of 0s.
            char [] gene = g.ToCharArray ();          // converted to an array.

            // a persons gender is set according parameter.
            char gender = '0';                      // assume this will be "F".

            if ( setSex == "M" )                   // set this person's gender.
                gender = '1';                      // select either 0=F or 1=M.

            if ( setSex == "F" )                   // set this person's gender.
                gender = '0';                      // select either 0=F or 1=M.

            // bits in gene are counted from right to left as n - 0.  Insert
            // the sex bit into the lsb position.
            gene [ GENE_LENGTH - 1 ] = gender;        // insert sex bit in lsb.

            return CharArrayToString ( gene );
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method returns a decimal version of Binary Number converted
        /// into decimal.  The function assumes the binary string to be an
        /// unsigned value.  Format is msb....lsb.  Conversion for values
        /// greater than 24-bits is undefined.  Exponent is the working bit
        /// width of the number to be converted.
        /// </summary>
        public int binStrToDec ( String binary, int exponent )
        {
            int result = 0;                              // the working result.
            int tmpExp = exponent;  // the minimum output word bit-width width.
            int i = 0;                                            // A counter.

            char [] binSample = binary.ToCharArray ();  // convert to an array.
            do
            {
                --tmpExp;

                if ( binSample [ i ] == '1' )
                    result += ( int ) ( Math.Pow ( 2.0D, ( double ) tmpExp ) );

            } while ( ++i < exponent );

            return result;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method calculates the standard deviations among the
        /// values in an array.
        /// </summary>
        public double calculateStdDev ( int [] array, double mean )
        {

            double sigma = 0.0D;

            for ( int counter = 0; counter < array.Length; counter++ )
            {
                sigma += Math.Pow ( ( ( ( double ) array [ counter ] ) - mean ),
                         2.0D );
            }

            double variance = sigma / ( ( double ) array.Length );  // get var.

            double stdDev = Math.Sqrt ( variance ); // take standard deviation.

            return stdDev;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method takes an array parameter and returns the sum of
        /// the data.
        /// </summary>
        public int calculateSum ( int [] array )
        {

            int sum = 0;

            for ( int i = 0; i < array.Length; i++ )      // loop thru and sum.
            {
                array [ i ] = i;
                sum += array [ i ];
            }

            return sum;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method converts an array of type char to a String.
        /// </summary>
        private String CharArrayToString ( char [] array )
        {
            String result = String.Empty;

            foreach ( char c in array )
            {
                result += c.ToString ();
            }

            return result;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method returns either true or false, pseudo-randomly.
        /// </summary>
        public bool coinFlip ()
        {
            return ( ( coin.NextDouble () > 0.49D ) ? true : false );
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method returns a new chromosome "child" as a result of the
        /// process of genetic cross-over.  The child chromosome is the same
        /// length as its parents.  Alleles from the parents are site-wise
        /// selected, at random, to construct the child.  Finally, the child
        /// chromosome is returned.  ASSUMPTION:  parent_1 and parent_2 have
        /// equi-length chomosomes.  Thus, the child (returned product) will
        /// have a length equal the parent(s).  Unequal length operations are
        /// undefined.
        /// </summary>
        public String crossover ( String parent_1, String parent_2 )
        {
            char [] child = new char [ parent_1.Length ];        // the result.
            char [] p_1 = parent_1.ToCharArray ();             // working copy.
            char [] p_2 = parent_2.ToCharArray ();             // working copy.

            for ( int i = 0; i < child.Length; i++ )
            {
                // select one parent or the other.
                if ( coinFlip () )
                    child [ i ] = p_1 [ i ];          // swap genetic material.
                else
                    child [ i ] = p_2 [ i ];          // swap genetic material.
            }

            return CharArrayToString ( child );       // return the chromosome.
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method returns a value in the output range y of the Gauss
        /// function if a value x lies in the function domain and is the
        /// provided parameter.
        /// </summary>
        public int gaussianFunction ( int x, double mean, double stdDev )
        {

            int result = 0;  // boundaries need control due to rounding errors.
            if ( ( x >= 255 ) || ( x <= 0 ) )
                result = x;                           // these are hard limits.
            else                     // all other values need to be calculated.
            {
                double sample = ( double ) x;  // convert int sample to double.

                double heightOfBell = 21000.0D;     // used to shape the curve.
                double widthOfBell = .55D;  // empirically-graphically defined.
                double bellFlanks = .39D;                                       // ditto.
                int fineOffset = 2;                                          // ditto.

                double numerator = -Math.Pow ( ( ( double ) x - mean ), 2.0D );
                double denominator = 2.0D *
                                   Math.Pow ( ( widthOfBell * stdDev ), 2.0D );
                double exponent = numerator / denominator;
                double exponential = Math.Pow ( Math.E, exponent );
                double fraction = 1.0D / ( stdDev *
                                  ( Math.Sqrt ( bellFlanks * Math.PI ) ) );

                result = ( ( int ) ( heightOfBell *
                                     fraction * exponential ) ) - fineOffset;
            }

            return result;                                // the final product.
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method creates and returns one chromosome of GENE_COUNT
        /// length.  A chromosome is an n-gene structure composed of
        /// concatenated "zero" and "one" (alleles).  The minimum length
        /// chromosome has one GENE.  The maximum length (multi-gene) is
        /// undefined.  Sexual phenotype is randomly assigned.
        /// </summary>
        public String makeChromosome ()
        {
            String chromosome = "";                              // chromosome.

            for ( int i = 0; i < ( GENE_COUNT - 1 ); i++ )
                chromosome += makeGene ( GAUSSIAN_TYPE );

            chromosome += assignPhenotype (); //append sexual phenotype scheme.

            return chromosome;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method creates and returns one chromosome of GENE_COUNT
        /// length.  A chromosome is an n-gene structure composed of
        /// concatenated "zero" and "one" (alleles).  The minimum length
        /// chromosome has one GENE.  The maximum length (multi-gene) is
        /// undefined.  Sexual phenotype is assigned by parameter.
        /// </summary>
        public String makeChromosome ( String setSex )
        {
            String chromosome = "";                              // chromosome.

            for ( int i = 0; i < ( GENE_COUNT - 1 ); i++ )
                chromosome += makeGene ( GAUSSIAN_TYPE );

            chromosome += assignPhenotype ( setSex );  // assign sex phenotype.

            return chromosome;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method directs the construction and return of an array of
        /// 256, 8-bit, integer encoded, Gaussian Normal distributed values.
        /// When returned, the array data is in bell curve configuration.
        /// </summary>
        public int [] makeGaussianDistribution ()
        {
            // calculate 2 ^ GENE_LENGTH.
            int geneMaxValue = ( int ) ( Math.Pow ( 2, GENE_LENGTH ) );
            int [] array = new int [ geneMaxValue ];  //array holds everything.

            for ( int i = 0; i < geneMaxValue; i++ )
                array [ i ] = signedByte ( i );       // integers -128 to +127.

            // sum the values in the array.
            double sum = ( double ) calculateSum ( array );
            // calculate the mean.
            double mean = sum / ( ( double ) geneMaxValue );
            // take the Std Deviation.
            double stdDev = calculateStdDev ( array, mean );

            // calculate.
            for ( int i = 0; i < geneMaxValue; i++ )
                if ( i == ( geneMaxValue / 2 ) )
                    array [ i ] = geneMaxValue - 1;
                else if ( i == ( geneMaxValue - 1 ) )
                    array [ i ] = 0;
                else
                    array [ i ] =
                              gaussianFunction ( array [ i ], mean, stdDev );

            return array;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method creates and returns one n-character Gene string
        /// composed of concatenated "zero" and "one" allele characters.
        /// The parameter, randomType, determines if the Gene returned is
        /// generated from the uniform random pool of integer base values
        /// or is based on a Gaussian distribution.
        /// </summary>
        public String makeGene ( int randomType )
        {
            // largest value a gene may take plus one.  Calculated as 2 ^ n.
            int geneMaxValue = ( int ) ( Math.Pow ( 2, GENE_LENGTH ) );

            // use Gaussian distribution made during instantiation.
            int geneValue = 0;      // individual sample from the distribution.
            switch ( randomType )
            {
                case 1:    // select value from a Gaussian normal distribution.
                    geneValue =
                    gaussDistribution [ coin.Next ( geneMaxValue ) ];
                    break;
                default:   // select a gene from a uniform random distribution.
                    geneValue =
                    coin.Next ( geneMaxValue );
                    break;
            }

            String gene = String.Empty;                       // output result.

            gene = Convert.ToString ( geneValue, 2 ); // convert to binary str.
            String outBuffer = new String ( '0', GENE_LENGTH - gene.Length );
            outBuffer += gene;

            return outBuffer;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method will induce a random mutation, in a single allele site,
        /// in the chromosome parameter provided.  Set by constant, the rate of
        /// MUTATION_PROBABILITY determines if the mutation occurs.
        /// </summary>
        public String randomMutation ( String chromosome )
        {
            ///////////// ADJUST THE RATE OF MUTATION ACCORDINGLY /////////////
            // per Nachman, 2000
            // the rate of mutation should be approximately
            // 2.5 X 10 ^ -8 per nucleotide per generation.
            // per Pennisi, 2003
            // there are less than 24.5K genes in humans.
            // 
            // ASSUME: chipanzees have somewhat fewer genes than do humans.
            //
            // Thus, a model having only 10 genes (or 2.45K times fewer
            // genes) might have a final rate of mutation probability
            // equivalent to: 0.0000245 or 0.00245%.
            // Rate = ( 24.5K / 10 ) * ( 1 / 100,000,000 )
            //
            const int MUTATION_RATE = 1;   // using a 1 in 40.816K SEARCH_SPACE
            const int SEARCH_SPACE = 40816; // a final probability is 0.00245%.
            ///////////// ADJUST THE RATE OF MUTATION ACCORDINGLY /////////////

            char [] candidate = chromosome.ToCharArray ();//mutation candidate.

            int mutationNumber = coin.Next ( SEARCH_SPACE );   // from samples.
            if ( mutationNumber < MUTATION_RATE )   //check if mutation occurs.
            {
                // pick one allele site.
                int mutationSite = coin.Next ( chromosome.Length );
                // select either 0 or 1.
                char mutantAllele = ( coinFlip () ? '0' : '1' );
                // replace allele at allele site with mutant allele.
                candidate [ mutationSite ] = mutantAllele;
            }

            // may (or -most likely- may not) contain mutation.
            return CharArrayToString ( candidate );
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method directs first the crossover reproduction then the
        /// single allele site mutation of an offspring chromosome.
        /// </summary>
        public String reproduction ( String parent_1, String parent_2 )
        {
            String child = String.Empty;

            child = crossover ( parent_1, parent_2 ); // do parental crossover.
            child = randomMutation ( child );       // do single-site mutation.

            return child;                   // return the offspring chromosome.
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method returns an 8-bit signed number between -128 and +127
        /// (inclusive) when provided with an integer valued 0 to 255.
        /// Parameter values greater than 255 or smaller than 0 are truncated.
        /// Examples are in=0 out=-128, in=127 out=-1, in=128 out=0, and
        /// in=255 out=127.
        /// </summary>
        public int signedByte ( int number )
        {

            int dummy = 0;

            if ( number > 255 )
                number = 255;                                    // hard limit.
            if ( number < 0 )
                number = 0;                                      // hard limit.

            if ( ( number >= 128 ) && ( number <= 255 ) )
                dummy = number - 128;

            if ( ( number >= 0 ) && ( number <= 127 ) )
                dummy = -Math.Abs ( number - 128 );

            return dummy;
        }

        #endregion Methods

        //
        //*********************************************************************
        //

        #region Getters & Setters

        /// <summary>
        /// this method tests and returns the gender of the
        /// chromosome provided.
        /// </summary>
        public String getGender ( String chromosome )
        {
            // by convention here 0 = female and 1 = male.
            if ( chromosome [ chromosome.Length - 1 ] == '1' )
                return "M";
            else
                return "F";
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method returns the gene found in the position along the
        /// length of a chromosome. Genes are numbered from right to left
        /// with the smallest numbered position on the right and the largest
        /// numbered position on the left.  Genes are numbered n, n-1, n-2,
        /// ... 2, 1, 0  etc.
        /// </summary>
        public String getGene ( String chromosome, int geneNumber )
        {

            int startAllele = ( GENE_COUNT * GENE_LENGTH ) -
                                ( ( geneNumber * GENE_LENGTH ) + GENE_LENGTH );
            int endAllele = startAllele + GENE_LENGTH;

            //return chromosome.Substring ( startAllele, endAllele );// Java
            return chromosome.Substring ( startAllele, GENE_LENGTH );// C#
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method returns the integer value of the gene, in the position
        /// requested, in chromosome provided.  Uses getGene to retreive the
        /// Allele sequence under examination.
        /// </summary>
        public int getGeneValue ( String chromosome, int geneNumber )
        {
            return binStrToDec ( getGene ( chromosome, geneNumber ),
                                 GENE_LENGTH );
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method does something.
        /// </summary>
        private void Template ()
        {
            // add code here.
        }

        #endregion Getters & Setters

        //
        //*********************************************************************
        //

    }
}
