﻿//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) MS Developers Forum                                         *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      TopForm.cs.                                                      *
//* CLASS    MsgInterface.cs. (Message Interface)                             *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     11/17/2013                                                       *
//* REVISION X.0.0.8                                                          *
//* BYTES    10,832                                                           *
//* TIME     19:00                                                            *
//* PURPOSE  To facilitate the orderly entry of agents into the model.        *
//* METHOD   By means of software simulation, model proto-humans.             *
//* DETAILS  Created in C# to model proto-humans.  Agents can be added to     *
//*          the model in groups of males and females.  This Class is built   *
//*          from MS Developers Forum Client-Server example source.           *
//* ENVIRON  Developed on a 64-bit Intel Core2Quad, 2.33GHz, with 4GB RAM.    *
//*          The Zotac m-ITX motherboard had board-based Nvidia 9300 graphics *
//*          acceleration. The Operating System was Microsoft (MS)            *
//*          Windows 8 Pro.  The MS Integrated Development Environment (IDE)  *
//*          MS Visual Studio 2010 Version 10.0.40219.1 SP1Rel and MS .NET    *
//*          Framework Version 4.0.30319 SP1Rel.                              *
//*                                                                           *
//*****************************************************************************
//

using System;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text;

namespace Nursery
{

    //
    //*************************************************************************
    // StateObject

    // Internal structure used in the interface.
    public class StateObject
    {

        public Socket workSocket = null;                 // socket socket.
        public const int BufferSize = 2048;              // Size of rcv buffer.
        public byte[] buffer = new byte[BufferSize];     // Rcv buffer.
        public StringBuilder sb = new StringBuilder();   // Rcv data string.

    }

    //
    //*************************************************************************
    // MsgInterface

    public class MsgInterface
    {
        #region Variables Dictionary

        private const String IPA = "10.0.0.13"; // the targeted Midwife's IPA.
        private const Int32 PORT = 1013;       // the targeted Midwife's port.

        // ManualResetEvent instances signal completion.
        private static ManualResetEvent connectDone =
                                        new ManualResetEvent(false);
        private static ManualResetEvent sendDone =
                                        new ManualResetEvent(false);
        private static ManualResetEvent receiveDone =
                                        new ManualResetEvent(false);

        private static IPAddress ipAddress = null;
        private static IPEndPoint remoteEP = null;

        // the response from the remote device.
        private static String response = String.Empty;
        // a network socket shared by the Class for sending and receiving
        private static Socket client = null;

        #endregion Variables Dictionary

        //
        //*********************************************************************
        //

        #region Methods

        //
        //*********************************************************************
        //

        /// <summary>
        /// this receives data from the network buffer and returns an object.
        /// </summary>

        public string Receive()
        {

            try
            {

                // Create the state object.
                StateObject state = new StateObject();
                state.workSocket = client;

                // Begin receiving the data from the remote device.
                client.BeginReceive(state.buffer, 0,
                                    StateObject.BufferSize, 0,
                                    new AsyncCallback(ReceiveCallback), state);

            }
            catch (Exception e)
            {

                Console.WriteLine(e.ToString());

            }
            receiveDone.WaitOne();

            return response;

        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method connects the socket to the server.
        /// </summary>
        private void ReceiveCallback(IAsyncResult ar)
        {
            try
            {

                // Retrieve the state object and the socket socket 
                // from the asynchronous state object.
                StateObject state = (StateObject)ar.AsyncState;
                Socket localClient = state.workSocket;

                // Read data from the remote device.
                int bytesRead = localClient.EndReceive(ar);
                if (bytesRead > 0)
                {

                    // collate the data for retrieval.
                    state.sb.Append(Encoding.ASCII.GetString
                        (state.buffer, 0, bytesRead));
                    response = state.sb.ToString();
                    receiveDone.Set();

                }

            }
            catch (Exception e)
            {

                Console.WriteLine(e.ToString());

            }

        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this sends data from the network buffer.
        /// </summary>

        public Boolean Send(String message)
        {

            // Send data to the remote device.
            // Convert the string data to byte data using ASCII encoding.
            byte[] byteData = Encoding.ASCII.GetBytes(message);

            // Begin sending the data to the remote device.
            client.BeginSend(byteData, 0, byteData.Length, 0,
                new AsyncCallback(SendCallback), client);

            sendDone.WaitOne();

            return true;

        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method connects the socket to the server.
        /// </summary>
        private void SendCallback(IAsyncResult ar)
        {

            try
            {

                // Retrieve the socket from the state object.
                Socket client = (Socket)ar.AsyncState;

                // Complete sending the data to the remote device.
                int bytesSent = client.EndSend(ar);

                // Signal that all bytes have been sent.
                sendDone.Set();

            }
            catch (Exception e)
            {

                Console.WriteLine(e.ToString());

            }

        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this starts the network socket and interrupts.
        /// </summary>

        public void StartClient()
        {

            // Connect to a remote device.
            try
            {

                ipAddress = IPAddress.Parse(IPA);
                remoteEP = new IPEndPoint(ipAddress, PORT);

                // Create a TCP/IP socket.
                client = new Socket(AddressFamily.InterNetwork,
                SocketType.Stream, ProtocolType.Tcp);

                //AsyncCallback acb = new AsyncCallback(ConnectCallback);
                //                client.BeginConnect(remoteEP, acb, client);
                client.Connect(remoteEP);

                connectDone.WaitOne(1000);

            }
            catch (Exception e)
            {

                Console.WriteLine(e.ToString());

            }

        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this shutsdown the network buffers and closes the object.
        /// </summary>

        public void StopClient()
        {

            // Release the socket.
            client.Shutdown(SocketShutdown.Both);
            client.Close();

        }

        //
        //*********************************************************************
        //

        #endregion Methods

        //
        //*********************************************************************
        //

    }

}
